"""Base aggregator interface."""

import abc
from typing import Generic, Optional, Sequence, TypeVar

from .typehints import Result, T

AggregateClass = TypeVar("AggregateClass", covariant=True)


class Aggregator(Generic[AggregateClass, Result], abc.ABC):
    """Interface for aggregators.

    Aggregators must implement the :meth:`~stupidb.aggregator.Aggregator.query`
    method. Aggregators are tied to a specific kind of aggregation. See the
    :meth:`~stupidb.aggregatetypes.UnaryAggregate.prepare` method for how to
    provide a custom aggregator.

    See Also
    --------
    stupidb.associative.SegmentTree
    stupidb.navigation.NavigationAggregator

    """

    @abc.abstractmethod
    def __init__(self, arguments: Sequence[T], cls: AggregateClass) -> None:
        """Initialize an aggregator from `arguments` and `cls`."""

    @abc.abstractmethod
    def query(self, begin: int, end: int) -> Optional[Result]:
        """Query the aggregator over the range from `begin` to `end`."""



# -*- coding: utf-8 -*-

"""Tests for `stupidb` package."""

import builtins
import itertools
import operator
import statistics
from datetime import date, timedelta
from typing import (
    Any,
    Callable,
    FrozenSet,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
    Type,
    TypeVar,
)

import cytoolz as toolz
import pytest

from stupidb.aggregation import Window
from stupidb.api import (
    aggregate,
    count,
    cov_pop,
    cov_samp,
    cross_join,
    exists,
    group_by,
    inner_join,
    left_join,
    limit,
    max,
    mean,
    min,
    mutate,
    order_by,
    over,
    right_join,
    select,
    sift,
    stdev_pop,
    stdev_samp,
    sum,
    table,
    total,
    var_pop,
    var_samp,
)
from stupidb.core import Relation, Table
from stupidb.row import Row

from .conftest import assert_rowset_equal


def test_projection(rows):
    expected = [
        dict(z="a", c=1, d=2),
        dict(z="b", c=2, d=-1),
        dict(z="a", c=3, d=4),
        dict(z="a", c=4, d=-3),
        dict(z="a", c=1, d=-3),
        dict(z="b", c=2, d=-3),
        dict(z="b", c=3, d=-3),
    ]
    pipeline = table(rows) >> select(
        c=lambda r: r["a"], d=lambda r: r["b"], z=lambda r: r["z"]
    )
    result = list(pipeline)
    assert_rowset_equal(result, expected)


def test_selection(rows):
    expected = [
        dict(z="a", c=1, d=2),
        dict(z="a", c=1, d=-3),
        dict(z="a", c=4, d=-3),
        dict(z="a", c=3, d=4),
        dict(z="b", c=2, d=-1),
        dict(z="b", c=2, d=-3),
        dict(z="b", c=3, d=-3),
    ]
    selection = (
        table(rows)
        >> select(c=lambda r: r["a"], d=lambda r: r["b"], z=lambda r: r["z"])
        >> sift(lambda r: True)
    )
    assert_rowset_equal(selection, expected)


def test_group_by(rows):
    expected = [
        {"c": 1, "mean": -0.5, "total": -1, "z": "a"},
        {"c": 2, "mean": -2.0, "total": -4, "z": "b"},
        {"c": 3, "mean": 4.0, "total": 4, "z": "a"},
        {"c": 4, "mean": -3.0, "total": -3, "z": "a"},
        {"c": 3, "mean": -3.0, "total": -3, "z": "b"},
    ]
    gb = (
        table(rows)
        >> select(c=lambda r: r["a"], d=lambda r: r["b"], z=lambda r: r["z"])
        >> sift(lambda _: True)
        >> group_by(c=lambda r: r["c"], z=lambda r: r["z"])
        >> aggregate(total=sum(lambda r: r["d"]), mean=mean(lambda r: r["d"]))
    )
    result = list(gb)
    assert_rowset_equal(result, expected)


def subclasses(cls: Type) -> FrozenSet[Type]:
    classes = cls.__subclasses__()
    return (
        frozenset({cls})
        | frozenset(classes)
        | frozenset(itertools.chain.from_iterable(map(subclasses, classes)))
    )


@pytest.mark.parametrize(
    "cls",
    (cls for cls in subclasses(Relation) if cls is not Relation and cls is not Table),
)
def test_from_iterable_is_invalid(cls):
    with pytest.raises(AttributeError):
        cls.from_iterable([dict(a=1)])


def test_cross_join(left, right):
    join = (
        table(left)
        >> cross_join(table(right))
        >> select(left_z=lambda r: r.left["z"], right_z=lambda r: r.right["z"])
    )
    result = list(join)
    assert len(result) == len(left) * len(right)
    expected = [
        {"left_z": left["z"], "right_z": right["z"]}
        for left, right in list(itertools.product(left, right))
    ]
    assert len(expected) == len(result)
    assert_rowset_equal(result, expected)


def test_inner_join(left, right, con):
    join = (
        table(left)
        >> inner_join(
            table(right),
            lambda left, right: (
                left["z"] == "a" and right["z"] == "a" and left["a"] == right["a"]
            ),
        )
        >> select(
            left_a=lambda r: r.left["a"],
            right_a=lambda r: r.right["a"],
            right_z=lambda r: r.right["z"],
            left_z=lambda r: r.left["z"],
        )
    )
    result = list(join)
    query = """
SELECT
    t.a AS left_a,
    s.a AS right_a,
    s.z AS right_z,
    t.z AS left_z
FROM left t
INNER JOIN right s
ON t.z = 'a'
  AND s.z = 'a'
  AND t.a = s.a"""
    expected = [
        dict(left_a=left_a, right_a=right_a, right_z=right_z, left_z=left_z)
        for left_a, right_a, right_z, left_z in con.execute(query).fetchall()
    ]
    assert_rowset_equal(result, expected)


def test_left_join(left, right, con):
    join = (
        table(left)
        >> left_join(table(right), lambda left, right: left["z"] == right["z"])
        >> select(left_z=lambda r: r.left["z"], right_z=lambda r: r.right["z"])
    )
    result = list(join)
    query = """
SELECT t.z AS left_z, s.z AS right_z
FROM left t
LEFT OUTER JOIN right s
ON t.z = s.z"""
    expected = [
        dict(left_z=left_z, right_z=right_z)
        for left_z, right_z in con.execute(query).fetchall()
    ]
    assert_rowset_equal(result, expected)


def test_right_join(left, right, con):
    join = (
        table(left)
        >> right_join(table(right), lambda left, right: left["z"] == right["z"])
        >> select(left_z=lambda r: r.left["z"], right_z=lambda r: r.right["z"])
    )
    result = list(join)
    query = """
SELECT t.z AS left_z, s.z AS right_z
FROM right s
LEFT OUTER JOIN left t
ON t.z = s.z"""
    expected = [
        dict(left_z=left_z, right_z=right_z)
        for left_z, right_z in con.execute(query).fetchall()
    ]
    assert_rowset_equal(result, expected)


def test_semi_join():
    rows = [
        dict(z="a", a=1, b=2),
        dict(z="b", a=2, b=-1),
        dict(z="a", a=3, b=4),
        dict(z="a", a=4, b=-3),
        dict(z="a", a=1, b=-3),
        dict(z="b", a=2, b=-3),
        dict(z="b", a=3, b=-3),
    ]
    other_rows = [
        dict(z="a", a=4, b=-3),
        dict(z="a", a=1, b=-3),
        dict(z="b", a=2, b=-3),
        dict(z="b", a=3, b=-3),
    ]

    pipeline = table(rows) >> sift(
        lambda r: exists(table(other_rows) >> sift(lambda o: r["z"] == o["z"]))
    )
    result = list(pipeline)
    assert result == rows


def test_semi_join_not_all_rows_match():
    rows = [
        dict(z="a", a=1, b=2),
        dict(z="b", a=2, b=-1),
        dict(z="a", a=3, b=4),
        dict(z="a", a=4, b=-3),
        dict(z="a", a=1, b=-3),
        dict(z="b", a=2, b=-3),
        dict(z="b", a=3, b=-3),
    ]
    other_rows = [dict(z="b", a=2, b=-3), dict(z="b", a=3, b=-3)]

    pipeline = table(rows) >> sift(
        lambda r: exists(table(other_rows) >> sift(lambda o: r["z"] == o["z"]))
    )
    result = list(pipeline)
    expected = [row for row in rows if row["z"] == "b"]
    assert result == expected


@pytest.mark.parametrize("offset", range(4))  # type: ignore[misc]
@pytest.mark.parametrize("lim", range(4))  # type: ignore[misc]
def test_valid_limit(rows: Sequence[Mapping[str, Any]], offset: int, lim: int) -> None:
    pipeline = table(rows) >> limit(lim, offset=offset)
    assert list(pipeline) == rows[offset : offset + lim]


@pytest.mark.parametrize(  # type: ignore[misc]
    ("offset", "lim"),
    ((offset, lim) for lim in range(-2, 1) for offset in range(-2, 1) if offset or lim),
)
def test_invalid_limit(
    rows: Sequence[Mapping[str, Any]],
    offset: int,
    lim: int,
) -> None:
    with pytest.raises(ValueError):
        table(rows) >> limit(lim, offset=offset)


def test_right_shiftable(rows):
    pipeline = (
        table(rows)
        >> select(c=lambda r: r["a"], d=lambda r: r["b"], z=lambda r: r["z"])
        >> sift(lambda r: True)
        >> group_by(c=lambda r: r["c"], z=lambda r: r["z"])
        >> aggregate(
            total=sum(lambda r: r["d"]),
            mean=mean(lambda r: r["d"]),
            my_samp_cov=cov_samp(lambda r: r["d"], lambda r: r["d"]),
            my_pop_cov=cov_pop(lambda r: r["d"], lambda r: r["d"]),
        )
    )

    expected = [
        {
            "c": 1,
            "mean": -0.5,
            "my_samp_cov": 15.0,
            "my_pop_cov": 7.5,
            "total": -1,
            "z": "a",
        },
        {
            "c": 2,
            "mean": -2.0,
            "my_samp_cov": 3.0,
            "my_pop_cov": 1.5,
            "total": -4,
            "z": "b",
        },
        {
            "c": 3,
            "mean": 4.0,
            "my_samp_cov": None,
            "my_pop_cov": 0.0,
            "total": 4,
            "z": "a",
        },
        {
            "c": 4,
            "mean": -3.0,
            "my_samp_cov": None,
            "my_pop_cov": 0.0,
            "total": -3,
            "z": "a",
        },
        {
            "c": 3,
            "mean": -3.0,
            "my_samp_cov": None,
            "my_pop_cov": 0.0,
            "total": -3,
            "z": "b",
        },
    ]
    result = list(pipeline)
    assert_rowset_equal(result, expected)


def test_rows_window(rows):
    pipeline = (
        table(rows)
        >> mutate(
            my_agg=sum(lambda r: r["a"])
            >> over(
                Window.rows(
                    order_by=[lambda r: r["e"]],
                    partition_by=[lambda r: r["z"]],
                    preceding=lambda r: 2,
                    following=lambda r: 0,
                )
            )
        )
        >> order_by(lambda r: r["z"], lambda r: r["e"])
    )
    result = list(pipeline)
    expected_aggrows = [
        {"my_agg": 1},
        {"my_agg": 2},
        {"my_agg": 4},
        {"my_agg": 8},
        {"my_agg": 8},
        {"my_agg": 4},
        {"my_agg": 7},
    ]
    expected = sorted(
        map(toolz.merge, rows, expected_aggrows),
        key=lambda r: (r["z"], r["e"]),
    )
    assert len(result) == len(expected)
    assert_rowset_equal(result, expected)


def test_rows_window_partition(rows):
    pipeline = (
        table(rows)
        >> mutate(
            my_agg=sum(lambda r: r.a) >> over(Window.rows(partition_by=[lambda r: r.z]))
        )
        >> order_by(lambda r: r.z, lambda r: r.e)
    )
    result = list(pipeline)
    expected_aggrows = [
        {"my_agg": 9},
        {"my_agg": 7},
        {"my_agg": 9},
        {"my_agg": 9},
        {"my_agg": 9},
        {"my_agg": 7},
        {"my_agg": 7},
    ]
    assert len(result) == len(rows)
    assert len(result) == len(expected_aggrows)
    expected = sorted(
        map(toolz.merge, rows, expected_aggrows),
        key=lambda r: (r["z"], r["e"]),
    )
    assert len(result) == len(expected)
    assert_rowset_equal(result, expected)


def test_range_window_with_multiple_ordering_keys_fails(rows):
    with pytest.raises(ValueError):
        table(rows) >> mutate(
            my_agg=sum(lambda r: r["a"])
            >> over(
                Window.range(
                    order_by=[lambda r: r["e"], lambda r: r["a"]],
                    partition_by=[lambda r: r["z"]],
                    preceding=lambda r: 2,
                    following=lambda r: 0,
                )
            )
        )


def test_range_window(rows):
    pipeline = (
        table(rows)
        >> mutate(
            my_agg=sum(lambda r: r["a"])
            >> over(
                Window.range(
                    order_by=[lambda r: r["e"]],
                    partition_by=[lambda r: r["z"]],
                    preceding=lambda r: 2,
                    following=lambda r: 0,
                )
            )
        )
        >> order_by(lambda r: r["z"], lambda r: r["e"])
        >> select(
            a=lambda r: r.a,
            e=lambda r: r.e,
            my_agg=lambda r: r.my_agg,
            z=lambda r: r.z,
        )
    )
    result = list(pipeline)
    expected_a = [
        {"a": 1, "e": 1, "my_agg": 1, "z": "a"},
        {"a": 3, "e": 3, "my_agg": 4, "z": "a"},
        {"a": 4, "e": 4, "my_agg": 7, "z": "a"},
        {"a": 1, "e": 5, "my_agg": 8, "z": "a"},
    ]
    expected_b = [
        {"a": 2, "e": 2, "my_agg": 2, "z": "b"},
        {"a": 2, "e": 6, "my_agg": 2, "z": "b"},
        {"a": 3, "e": 7, "my_agg": 5, "z": "b"},
    ]

    expected = expected_a + expected_b

    assert len(result) == len(expected)
    assert_rowset_equal(result, expected)


def test_temporal_range_window(t_table, t_rows):
    query = t_table >> mutate(
        avg_balance=mean(lambda r: r["balance"])
        >> over(
            Window.range(
                order_by=[lambda r: r["date"]],
                partition_by=[lambda r: r["name"]],
                preceding=lambda r: timedelta(days=3),
                following=lambda r: timedelta(days=0),
            )
        )
    )
    result = list(query)
    expected = [
        {
            "avg_balance": 2.0,
            "balance": 2,
            "date": date(2018, 1, 1),
            "name": "alice",
        },
        {
            "avg_balance": 3.0,
            "balance": 4,
            "date": date(2018, 1, 4),
            "name": "alice",
        },
        {
            "avg_balance": 0.5,
            "balance": -3,
            "date": date(2018, 1, 6),
            "name": "alice",
        },
        {
            "avg_balance": -0.666_666_666_666_666_6,
            "balance": -3,
            "date": date(2018, 1, 7),
            "name": "alice",
        },
        {
            "avg_balance": -1.0,
            "balance": -1,
            "date": date(2018, 1, 2),
            "name": "bob",
        },
        {
            "avg_balance": -2.0,
            "balance": -3,
            "date": date(2018, 1, 3),
            "name": "bob",
        },
        {
            "avg_balance": -2.333_333_333_333_333_5,
            "balance": -3,
            "date": date(2018, 1, 4),
            "name": "bob",
        },
    ]
    assert_rowset_equal(result, expected)


def test_agg(rows):
    pipeline = table(rows) >> aggregate(
        sum=sum(lambda r: r["e"]),
        mean=mean(lambda r: r["e"]),
        count=count(lambda r: r["e"]),
    )
    (result,) = list(pipeline)
    assert result["sum"] == 28
    assert result["mean"] == result["sum"] / result["count"]


def test_invalid_agg(rows):
    with pytest.raises(TypeError, match="Invalid projection"):
        select(
            not_an_agg=lambda r: r["e"],
            my_agg=sum(lambda r: r["e"]),
            my_agg2=mean(lambda r: r["e"]),
            my_count=count(lambda r: r["e"]),
        )


T = TypeVar("T")
U = TypeVar("U")


def cumagg(seq: Iterable[T], combine: Callable[[T, T], T]) -> Iterator[T]:
    """Cumulative aggregation."""
    it = iter(seq)
    result = next(it)
    yield result
    for value in it:
        result = combine(value, result)
        yield result


def test_cumagg(rows):
    window = Window.rows(order_by=[lambda r: r.e])
    query = table(rows) >> select(
        cumsum=sum(lambda r: r.e) >> over(window),
        cumcount=count(lambda r: r.e) >> over(window),
    )
    result = list(query)
    expected = [
        dict(cumsum=cs, cumcount=cc)
        for cs, cc in zip(cumagg(range(1, 8), operator.add), range(1, 8))
    ]
    assert_rowset_equal(result, expected)


def test_total_vs_sum():
    rows = [dict(value=None), dict(value=None)]
    query = table(rows) >> aggregate(
        sum=sum(lambda r: r.value), total=total(lambda r: r.value)
    )
    (result_row,) = list(query)
    assert result_row.sum is None
    assert result_row.total == 0


def test_minmax(rows):
    query = table(rows) >> aggregate(min=min(lambda r: r.e), max=max(lambda r: r.e))
    result = list(query)
    e = [r["e"] for r in rows]
    expected = [dict(min=builtins.min(e), max=builtins.max(e))]
    assert_rowset_equal(result, expected)


def test_min_max_window(t_rows):
    window = Window.range(partition_by=[lambda r: r.name])
    query = table(t_rows) >> select(
        min_date=min(lambda r: r.date) >> over(window),
        max_date=max(lambda r: r.date) >> over(window),
    )
    result = list(query)
    expected = [
        dict(min_date=date(2018, 1, 1), max_date=date(2018, 1, 7)),
        dict(min_date=date(2018, 1, 1), max_date=date(2018, 1, 7)),
        dict(min_date=date(2018, 1, 1), max_date=date(2018, 1, 7)),
        dict(min_date=date(2018, 1, 1), max_date=date(2018, 1, 7)),
        dict(min_date=date(2018, 1, 2), max_date=date(2018, 1, 4)),
        dict(min_date=date(2018, 1, 2), max_date=date(2018, 1, 4)),
        dict(min_date=date(2018, 1, 2), max_date=date(2018, 1, 4)),
    ]
    assert_rowset_equal(result, expected)


def test_variance_window(t_rows):
    window = Window.range(partition_by=[lambda r: r.name])
    query = table(t_rows) >> select(
        name=lambda r: r.name,
        var=var_samp(lambda r: r.balance) >> over(window),
        std=stdev_samp(lambda r: r.balance) >> over(window),
        popvar=var_pop(lambda r: r.balance) >> over(window),
        popstd=stdev_pop(lambda r: r.balance) >> over(window),
    )
    result = list(query)
    alice = [row["balance"] for row in t_rows if row["name"] == "alice"]
    bob = [row["balance"] for row in t_rows if row["name"] == "bob"]
    expected = {
        Row(
            dict(
                name="alice",
                var=statistics.variance(alice),
                std=statistics.stdev(alice),
                popvar=statistics.pvariance(alice),
                popstd=statistics.pstdev(alice),
            ),
            _id=1,
        ),
        Row(
            dict(
                name="bob",
                var=statistics.variance(bob),
                std=statistics.stdev(bob),
                popvar=statistics.pvariance(bob),
                popstd=statistics.pstdev(bob),
            ),
            _id=2,
        ),
    }
    assert set(result) == expected



"""Abstract and concrete aggregation types."""

from __future__ import annotations

import abc
from typing import Callable, ClassVar, Generic, Sequence, Tuple

from .aggregator import Aggregator
from .row import AbstractRow
from .typehints import Getter, Output


class Aggregate(Generic[Output], abc.ABC):
    """An aggregate or window function."""

    __slots__ = ()
    aggregator_class: ClassVar[Callable[..., Aggregator]]

    @classmethod
    def prepare(
        cls,
        possible_peers: Sequence[Tuple[int, AbstractRow]],
        getters: Tuple[Getter, ...],
        order_by_columns: Sequence[str],
    ) -> Aggregator[Aggregate[Output], Output]:
        """Prepare an aggregation of this type for computation."""
        arguments = [
            tuple(getter(peer) for getter in getters) for _, peer in possible_peers
        ]
        return cls.aggregator_class(arguments, cls)



from datetime import date

from stupidb.api import Window, first, lag, last, lead, nth, over, select, table

from .conftest import assert_rowset_equal


def test_first_last(t_rows):
    window = Window.range(partition_by=[lambda r: r.name])
    query = table(t_rows) >> select(
        first_date=first(lambda r: r.date) >> over(window),
        last_date=last(lambda r: r.date) >> over(window),
        first_date_nulls=first(lambda r: None) >> over(window),
    )
    result = list(query)
    expected = [
        dict(
            first_date=date(2018, 1, 1),
            last_date=date(2018, 1, 7),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 1),
            last_date=date(2018, 1, 7),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 1),
            last_date=date(2018, 1, 7),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 1),
            last_date=date(2018, 1, 7),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 2),
            last_date=date(2018, 1, 4),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 2),
            last_date=date(2018, 1, 4),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 2),
            last_date=date(2018, 1, 4),
            first_date_nulls=None,
        ),
    ]
    assert_rowset_equal(result, expected)


def test_nth(t_rows):
    query = table(t_rows) >> select(
        nth_date=nth(lambda r: r.date, lambda r: 1)
        >> over(Window.range(partition_by=[lambda r: r.name]))
    )
    result = list(query)
    expected = [
        dict(nth_date=date(2018, 1, 4)),
        dict(nth_date=date(2018, 1, 4)),
        dict(nth_date=date(2018, 1, 4)),
        dict(nth_date=date(2018, 1, 4)),
        dict(nth_date=date(2018, 1, 3)),
        dict(nth_date=date(2018, 1, 3)),
        dict(nth_date=date(2018, 1, 3)),
    ]
    assert_rowset_equal(result, expected)


def test_nth_past_frame(t_rows):
    query = table(t_rows) >> select(
        nth_date=nth(lambda r: r.date, lambda r: 4000)
        >> over(Window.range(partition_by=[lambda r: r.name]))
    )
    result = list(query)
    expected = [
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
    ]
    assert_rowset_equal(result, expected)


def test_nth_past_frame_preceding_following(t_rows):
    query = table(t_rows) >> select(
        nth_date=nth(lambda r: r.date, lambda r: 4000)
        >> over(
            Window.range(
                partition_by=[lambda r: r.name],
                preceding=lambda r: 200,
                following=lambda r: 1000,
            )
        )
    )
    result = list(query)
    expected = [
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
    ]
    assert_rowset_equal(result, expected)


def test_lead_lag(t_rows):
    window = Window.range(partition_by=[lambda r: r.name])
    query = table(t_rows) >> select(
        lead_date=lead(lambda r: r.date, lambda r: 1) >> over(window),
        lag_date=lag(lambda r: r.date, lambda r: 1) >> over(window),
    )
    result = list(query)
    expected = [
        dict(lead_date=date(2018, 1, 4), lag_date=None),
        dict(lead_date=date(2018, 1, 6), lag_date=date(2018, 1, 1)),
        dict(lead_date=date(2018, 1, 7), lag_date=date(2018, 1, 4)),
        dict(lead_date=None, lag_date=date(2018, 1, 6)),
        dict(lead_date=date(2018, 1, 3), lag_date=None),
        dict(lead_date=date(2018, 1, 4), lag_date=date(2018, 1, 2)),
        dict(lead_date=None, lag_date=date(2018, 1, 3)),
    ]
    assert_rowset_equal(result, expected)



import pytest

from stupidb.reversed import Reversed


@pytest.fixture
def rev():
    return Reversed([1, 2, 3, 4])


def test_iter(rev):
    assert list(rev) == list(reversed(rev.values))


def test_invertible(rev):
    assert list(reversed(rev)) == rev.values


@pytest.mark.parametrize(("index", "expected"), [(0, 4), (1, 3), (2, 2), (3, 1)])
def test_indexing(rev, index, expected):
    assert rev[index] == expected


def test_invalid_indexing(rev):
    with pytest.raises(IndexError):
        rev[5]


@pytest.mark.parametrize(("index", "expected"), [(-1, 1), (-2, 2), (-3, 3), (-4, 4)])
def test_negative_indexing(rev, index, expected):
    assert rev[index] == expected


def test_invalid_negative_indexing(rev):
    with pytest.raises(IndexError):
        rev[-5]


def test_repr(rev):
    assert repr(rev) == "Reversed([4, 3, 2, 1])"



"""Implementation of an indexable :class:`stupidb.reversed.Reversed` object."""

from typing import Any, Sequence, TypeVar

T = TypeVar("T", covariant=True)


class Reversed(Sequence[T]):
    """A sequence that iterates backwards over its input."""

    __slots__ = ("values",)

    def __init__(self, values: Sequence[T] = ()) -> None:
        self.values = values

    def __repr__(self) -> str:
        return f"{type(self).__name__}({list(self)!r})"

    def __len__(self) -> int:
        return len(self.values)

    def __getitem__(self, index: Any) -> Any:
        # Why must this be Any -> Any?
        nvalues = len(self)
        if -nvalues <= index < nvalues:
            offset = nvalues * (index >= 0) - index - 1
            return self.values[offset]
        raise IndexError(index)



from stupidb.indextree import IndexTree


def test_tree_repr_fanout_2():
    tree = IndexTree(height=3, fanout=2)
    expected = """\
|-- 0
    |-- 1
        |-- 3
        |-- 4
    |-- 2
        |-- 5
        |-- 6"""
    assert repr(tree) == expected


def test_tree_repr_fanout_4():
    tree = IndexTree(height=2, fanout=4)
    expected = """\
|-- 0
    |-- 1
    |-- 2
    |-- 3
    |-- 4"""
    assert repr(tree) == expected


def test_parent_of_root():
    tree = IndexTree(height=2, fanout=2)
    assert tree.parent(0) == 0
